;; map-compiler.lisp
;;
;; Copyright 2011 Jonathan L. Fischer <jonathan@mohiji.org>.  All rights reserved.
;;
;; Demo code to take a file with an s-expression that looks like this:
;;
;; (map
;;  (name "The Blocky River")
;;  (background "cloudy-sky.png")
;;  (dimensions 10 10)
;;  (light-color "#d1ae4d")
;;  (light-direction -0.1212 -0.4848 0.4848)
;;
;;  (layer 2 2 2 2 2 2 2 2 2 2
;;         2 2 2 2 2 2 2 2 2 2
;;         2 2 2 6 6 2 2 2 2 2
;;         2 2 2 6 6 2 2 2 2 2
;;         2 2 2 6 6 6 2 2 2 2
;;         2 2 2 2 6 6 2 2 2 2
;;         2 2 2 2 2 2 2 2 2 2
;;         2 2 2 2 2 2 2 2 2 2
;;         2 2 2 2 2 2 2 2 2 2
;;         2 2 2 2 2 2 2 2 2 2)
;;
;;  (layer 3 5 5 3 3 3 3 3 3 3
;;         3 5 5 3 3 3 3 3 3 3
;;         3 5 5 0 0 3 3 3 3 3
;;         3 5 5 0 0 3 3 3 3 3 
;;         3 5 5 0 0 0 3 3 3 3
;;         3 5 5 3 0 0 3 3 3 3
;;         3 5 5 3 3 3 3 3 3 3
;;         3 5 5 3 3 3 3 3 3 3
;;         3 5 5 3 3 3 3 3 3 3
;;         3 5 5 3 3 3 3 3 3 3))
;;
;; And turn it into a binary file easily loaded into C. (see
;; read-blockmap.c for how to do that.)

;; We'll be using Marijn Haverbeke's ieee-floats
;; (http://common-lisp.net/project/ieee-floats/).  You'll need
;; to download that library to use this, unless you just want
;; to leave out the parts that deal w/ floating point numbers.
(require 'ieee-floats)

;; Define some constants to help identify that these files are in fact
;; map files.
(defparameter *blockmap-magic* 123)
(defparameter *blockmap-version* 1)

;; Helper structure to hold the data.  Not really necessary, but it makes
;; the rest of the code cleaner
(defclass blockmap ()
  ((name        :accessor map-name)
   (background  :accessor map-background)
   (light-color :accessor map-light-color)
   (light-dir   :accessor map-light-dir)
   (width       :accessor map-width)
   (height      :accessor map-height)
   (layers      :accessor map-layers :initform (list))))

;; Parse a list of the above form out into a blockmap class
(defun parse-blockmap-list (l)
  (let ((map    (make-instance 'blockmap))
        (fields (cdr l)))
    (loop for item in fields do
         (let ((key (first item))
               (value (second item)))
           ;; This is pretty much just a switch statement.  The format
           ;; has each field as a list with the first value being the
           ;; field name/type.
           (cond ((eq key 'name)
                  (setf (map-name map) value))
                 ((eq key 'background)
                  (setf (map-background map) value))
                 ((eq key 'light-color)
                  (setf (map-light-color map) (cdr item)))
                 ((eq key 'light-direction)
                  (setf (map-light-dir map) (cdr item)))
                 ((eq key 'dimensions)
                  (setf (map-width map) (first (cdr item)))
                  (setf (map-height map) (second (cdr item))))
                 ((eq key 'layer)
                  (push (cdr item) (map-layers map)))
                 (t
                  (format t "Unknown key ~a~%" key)))))
    map))

;; Take a normal string and convert it into a 256 byte null-terminated
;; array.  (e.g. a fixed length C string)
(defun make-c-string (str)
  (assert (< (length str) 256))
  (let ((final-str (make-array 256 :element-type '(unsigned-byte 8) :initial-element 0)))
    (map-into final-str (lambda (x) (char-code x)) str)
    final-str))

;; Take a list of byte values (r g b) and turn it into a
;; packed 32 bit ARGB value.  (I'm using little endian here)
(defun make-argb (values)
  (assert (>= (length values) 3))
  (let ((result 0))
    (setf (ldb (byte 8 24) result) 255)
    (setf (ldb (byte 8 16) result) (first values))
    (setf (ldb (byte 8 8) result) (second values))
    (setf (ldb (byte 8 0) result) (third values))
    result))

;; Write a 32 bit (little endian) integer out as a set of bytes
(defun write-int32 (value stream)
  (write-byte (ldb (byte 8 0) value) stream)
  (write-byte (ldb (byte 8 8) value) stream)
  (write-byte (ldb (byte 8 16) value) stream)
  (write-byte (ldb (byte 8 24) value) stream))

;; Take a blockmap object and write it out to a file.
(defun write-blockmap (map filename)
  (with-open-file (f filename 
                     :direction :output
                     :element-type '(unsigned-byte 8)
                     :if-exists :supersede)
    ;; Write out the header        
    (write-byte *blockmap-magic* f)
    (write-byte *blockmap-version* f)
    
    ;; Then the title and background strings
    (write-sequence (make-c-string (map-name map)) f)
    (write-sequence (make-c-string (map-background map)) f)
    
    ;; Light color
    (write-int32 (make-argb (map-light-color map)) f)
    
    ;; Light direction
    (write-int32 (ieee-floats:encode-float32 (first (map-light-dir map))) f)
    (write-int32 (ieee-floats:encode-float32 (second (map-light-dir map))) f)
    (write-int32 (ieee-floats:encode-float32 (third (map-light-dir map))) f)

    ;; Layer information
    (write-byte (length (map-layers map)) f)
    (write-byte (map-width map) f)
    (write-byte (map-height map) f)

    ;; And the layers themselves
    (loop for layer in (map-layers map) do
         (let ((data (make-array (* (map-width map) (map-height map))
                                 :element-type '(unsigned-byte 8)
                                 :initial-contents layer)))
           (write-sequence data f)))))